#!/usr/bin/env python

import dronin_pyqtgraph as pg
from dronin_pyqtgraph.Qt import QtCore, QtGui
import numpy as np

from dronin.logviewer.plotdockarea import PlotDockArea

from dronin_pyqtgraph.dockarea import *

import six

def get_data_series(obj_name, fields):
    data = get_series(obj_name)

    base_fields = [ f.split(':')[0] for f in fields ]
    peeled = data[base_fields]

    if ':' not in ''.join(fields):
        try:
            return peeled.view(dtype='float').reshape(-1, 2)
        except Exception:
            pass

    rows = peeled.shape[0]
    cols = len(fields)
    outp = np.empty((peeled.shape[0], len(fields)))

    for j in range(cols):
        field_info = fields[j].split(':')

        if len(field_info) > 1:
            subs = int(field_info[1])
            for i in range(rows):
                val = peeled[i][j][subs]

                outp[i,j] = float(val)
        else:
            for i in range(rows):
                val = peeled[i][j]

                outp[i,j] = float(val)

    return outp

def add_plot_area(data_series, dock_name, axis_label, legend=False, **kwargs):
    dock = Dock(dock_name, size=(800, 300))

    pw = pg.PlotWidget()

    if legend:
        pw.addLegend()

    colors = [ 'w', 'm', 'y', 'c' ]
    idx = 0

    for plot_name, data in data_series.items():
        pw.plot(data, antialias=True, name='&nbsp;'+plot_name, pen=pg.mkPen(colors[idx]), **kwargs)
        idx += 1

    # pen=None, symbol='o', symbolSize=2.5

    pw.setLabel('left', axis_label)

    pw.setMenuEnabled(menus_enabled)

    dock.addWidget(pw)

    global last_plot

    if (last_plot is not None) and (last_plot.isVisible()):
        area.addDock(dock, 'bottom', last_plot)
    else:
        area.addDock(dock, 'left', dui)

    last_plot = dock

    return pw

def plot_vs_time(obj_name, fields):
    if not isinstance(fields, list):
        fields = [fields]

    if len(fields) > 1:
        left_axis_label = '%s<br>&nbsp;<br>&nbsp;' % (obj_name)
        legend = True
    else:
        left_axis_label = '%s<br>%s<br>%s' % (obj_name, fields[0], objtyps[obj_name]._units[fields[0].split(':')[0]])
        legend = False

    data_series = {}
    for f in fields:
        data_series[obj_name + '.' + f] = get_data_series(obj_name, ['time', f])

    global win_num

    win_num += 1
    dock_name = "TimeSeries%d" % (win_num)

    return add_plot_area(data_series, dock_name, left_axis_label, legend=legend)

def clear_plots(skip=None):
    containers, docks = area.findAll()

    for d in list(docks.values()):
        if skip is not None and d in skip:
            continue

        if (not d.name().startswith("TimeSeries")) and (d.name() != "Waiting..."):
            continue

        d.close()

def get_series(name):
    global series

    if name in series:
        return series[name]

    typ = t.uavo_defs.find_by_name(name)

    series[name] = t.as_numpy_array(typ)

    return series[name]

def scan_for_events(uv):
    flight_mode = -1
    armed = -1

    events = []

    for u in uv:
        if u._name == 'UAVO_FlightStatus':
            ev = []
            # u.Armed DISARMED/ARMING/ARMED
            # u.FlightMode

            if u.Armed != armed:
                armed = u.Armed

                ev.append(u.ENUMR_Armed[armed])

            if u.FlightMode != flight_mode:
                flight_mode = u.FlightMode

                ev.append('MODE:' + u.ENUMR_FlightMode[flight_mode])

            if len(ev):
                tup = (u.time, '/'.join(ev))
                events.append(tup)

    return events

def handle_open(ignored=False, fname=None):
    from dronin import telemetry, uavo

    if fname is None:
        fname = QtGui.QFileDialog.getOpenFileName(win, 'Open file', filter="Log files (*.drlog *.txt)")

    with pg.ProgressDialog("0 objects read...", wait=500, maximum=1000, cancelText=None) as dlg:
        global t

        if not isinstance(fname, six.string_types):
            fname = fname[0]

        f = open(fname, 'rb')
        num_bytes = 1000000000

        try:
            import os

            stat_info = os.fstat(f.fileno())

            num_bytes = stat_info.st_size
        except Exception:
            print("Couldn't stat file")
            pass

        def cb(n_objs, n_bytes):
            # Top out at 90%, so the dialog doesn't hang at 100%
            # during the non-reading operations...
            permille = (n_bytes * 900.0) / num_bytes

            # should not happen, but cover the case anyways
            if (permille > 900.0): permille = 900.0
            dlg.setValue(permille)
            dlg.setLabelText("%d objects read..." % n_objs)

        t = telemetry.FileTelemetry(f, parse_header=True, service_in_iter=True,
                    gcs_timestamps=None, name=fname, progress_callback=cb)

        global series, objtyps
        series = {}
        objtyps = {}

        for typ in t.uavo_defs.values():
            short_name = typ._name[5:]
            objtyps[short_name] = typ

        event_series = scan_for_events(t)

        global last_plot
        last_plot = None

        thrust_plot = plot_vs_time('StabilizationDesired', 'Thrust')

        clear_plots(skip=[last_plot])

        for tm,text in event_series:
            thrust_plot.addLine(x=tm)
            # label=text, labelOpts={'rotateAxis' : (1,0)} )

        dlg.setValue(925)
        plot_vs_time('AttitudeActual', ['Yaw', 'Roll', 'Pitch'])
        dlg.setValue(975)
        plot_vs_time('Gyros', ['x', 'y', 'z'])
        plot_vs_time('ActuatorCommand', ['Channel:0', 'Channel:1', 'Channel:2', 'Channel:3'])

        objtyps = { k:v for k,v in objtyps.items() if v in t.last_values }

        #add all non-settings objects, and autotune, to the keys.
        objSel.clear()

        for typnam in sorted(objtyps.keys()):
            if typnam == "SystemIdent" or not objtyps[typnam]._is_settings:
                objSel.addItem(typnam)

        objSel.setEnabled(True)

def updateItems(i):
    uavo = objtyps[str(objSel.currentText())]

    itemSel.clear()

    for i in range(0, len(uavo._num_subelems)):
        field_name = uavo._fields[3+i]
        num_subs = uavo._num_subelems[i]

        if num_subs == 1:
            itemSel.addItem(field_name)
        else:
            for j in range(0, num_subs):
                itemSel.addItem('%s:%d'%(field_name, j))

    itemSel.setEnabled(True)
    addPlotBtn.setEnabled(True)

def addTimeSeries():
    plot_vs_time(str(objSel.currentText()), str(itemSel.currentText()))

def enableContextuals(enabled):
    global menus_enabled

    menus_enabled = enabled

    containers, docks = area.findAll()

    for dock in docks.values():
        for w in dock.widgets:
            if isinstance(w, pg.widgets.PlotWidget.PlotWidget):
                w.setMenuEnabled(enabled)

# Create our application, consisting of dockarea stuff.
# For now fix the size at 1024x680, so that we can make it work well in this
# size
app = QtGui.QApplication([])
win = QtGui.QMainWindow()
area = PlotDockArea()

win.setCentralWidget(area)
win.resize(1024, 680)

menubar = win.menuBar()

win_num = 0
menus_enabled = False

openAction = QtGui.QAction("&Open", win)
openAction.setShortcut(QtGui.QKeySequence.Open)
openAction.triggered.connect(handle_open)

quitAction = QtGui.QAction('&Quit', win)
quitAction.setShortcut(QtGui.QKeySequence.Quit)
quitAction.triggered.connect(app.quit)

fileMenu = menubar.addMenu('&File')

fileMenu.addAction(openAction)
fileMenu.addAction(quitAction)

## Create docks, place them into the window.
dl = Dock("Waiting...", size=(800, 1))

dui = Dock("UI", size=(224, 300))

area.addDock(dui, 'right')     ## place d2 at right edge of dock area
area.addDock(dl, 'left')

dui.addWidget(QtGui.QLabel("Obj:"), 0, 0)

objSel = QtGui.QComboBox()
objSel.setEnabled(False)
dui.addWidget(objSel, 0, 1)

dui.addWidget(QtGui.QLabel("Item:"), 1, 0)

itemSel = QtGui.QComboBox()
itemSel.setEnabled(False)
addPlotBtn = QtGui.QPushButton("&Add time-series plot")
addPlotBtn.setEnabled(False)

contextualCb = QtGui.QCheckBox("Enable contextual menus")
contextualCb.setChecked(False)

dui.addWidget(itemSel, 1, 1)
dui.addWidget(addPlotBtn, 2, 0, 1, 2)
dui.layout.addItem(QtGui.QSpacerItem(20, 20, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding), 3, 0)
dui.layout.addWidget(contextualCb, 4, 0, 1, 2)
dui.layout.addItem(QtGui.QSpacerItem(20, 20, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding), 5, 0)

objSel.currentIndexChanged.connect(updateItems)
addPlotBtn.clicked.connect(addTimeSeries)
contextualCb.toggled.connect(enableContextuals)

win.show()

def main():
    import sys

    if len(sys.argv) == 2:
        handle_open(fname=sys.argv[1])

    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()



# Copyright (C) 2016 dRonin, http://dronin.org
# Licensed under the GNU LGPL version 2.1 or any later version (see COPYING.LESSER)

from struct import Struct
from datetime import datetime

from xml.etree import ElementTree

def _prettify(elem):
    """Return a pretty-printed XML string for the Element.
    """
    from xml.dom import minidom

    rough_string = ElementTree.tostring(elem, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="  ")

class LogFSImport(dict):
    # vastly varying arena_size -- 0x1000 to 0x20000 ... though 0x20000 (aq32) is
    # probably illegit   all slot sizes 0x100 for settings

    known_magics = {
            'aq32/quanton'                   : 0x3b1b14cf,
            'brain'                          : 0x3bb141cf,
            'cc3d'                           : 0x99abcdef,
            'discoveryf4'                    : 0x99abcfef,
            'sparky2 int/revo/sim'           : 0x99abcedf,
            'sparky2'                        : 0x77abcedf,
            'simulation'                     : 0x89abceef,
            'colibri'                        : 0x3bb141ed,
            'flyingf3/lux/naze/pipx/sparky'  : 0x9ae1ee11
            }

    # magic, state .. 8 byte offset in
    arena_header = Struct('II')

    arena_states = { 'ERASED' : 0xffffffff,
            'RESERVED'  : 0xe6e6ffff,
            'ACTIVE'    : 0xe6e66666,
            'OBSOLETE'  : 0x00000000 }

    # state, obj_id, obj_inst_id, obj_size
    slot_header = Struct('IIHH')

    slot_states = { 'EMPTY' : 0xffffffff,
            'RESERVED'      : 0xfafaffff,
            'ACTIVE'        : 0xfafaaaaa,
            'OBSOLETE'      : 0x00000000 }

    slot_size = 256

    def __init__(self, githash, contents):
        from dronin import uavo_collection

        our_magic = None

        pos = 0

        prev_state_good = False
        in_good_arena = False

        obj_offsets = {}

        while pos + self.slot_size < len(contents):
            # check if we're in a sane state

            # assumption: arena sizes are power of 2, at least 2048 bytes
            if (pos & 0x7ff) == 0:
                magic,arena_state = self.arena_header.unpack_from(contents, pos)

                if our_magic is None:
                    for k in self.known_magics:
                        if self.known_magics[k] == magic:
                            #print "Selected magic number for targ " + k
                            our_magic = magic

                if magic == our_magic:
                    # let's assume this is good.  the slot is in the right place to
                    # actually be an arena header, and it has the right magic.

                    # print "Found probable arena at %x, state=%08x" % (pos, arena_state)

                    if arena_state == self.arena_states['ACTIVE']:
                        in_good_arena = True
                    else:
                        in_good_arena = False

                    pos += self.slot_size
                    prev_state_good = True

                    continue

            state, obj_id, inst_id, size = self.slot_header.unpack_from(contents, pos)

            prev_state_good = False

            if state == self.slot_states['ACTIVE']:
                if in_good_arena:
                    obj_offsets[(obj_id, inst_id)] = pos + self.slot_header.size

                prev_state_good = True

            pos += self.slot_size

        # We do this in two passes; A) so we can try and guess a good version from the
        # set of IDs if it is not known in the future, B) to unpack all at once

        uavo_defs = uavo_collection.UAVOCollection()

        uavo_defs.from_git_hash(githash)

        self.githash = githash

        for id_tup,offset in obj_offsets.items():
            obj_id, inst_id = id_tup

            uavo_key = '{0:08x}'.format(obj_id)

            obj = uavo_defs.get(uavo_key)

            if obj is not None:
                # print "  slot state=%08x, objid=%08x, instid=%04x, size=%d"%(state, obj_id, inst_id, size)

                objInstance = obj.from_bytes(contents, 0, None, offset=offset)

                self[obj._name] = objInstance

    def ExportXML(self):
        top = ElementTree.Element('uavobjects')

        comment = ElementTree.Comment('Automatically generated by dronin.logfs')
        top.append(comment)
        comment = ElementTree.Comment('Exported at %s' % (datetime.utcnow()))
        top.append(comment)
        comment = ElementTree.Comment('Interpreted using githash %s' % (self.githash))
        top.append(comment)

        settings = ElementTree.SubElement(top, 'settings')

        for obj_name in sorted(self.keys()):
            exported_obj = self[obj_name].to_xml_elem()
            settings.append(exported_obj)

        return _prettify(top)



from dronin_pyqtgraph.dockarea import DockArea, Dock
import dronin_pyqtgraph as pg
from dronin_pyqtgraph.dockarea.Container import VContainer, TContainer

# This code is so we can link plots automatically when time series 
# are above each other...

class PlotDockArea(DockArea):
    def addDock(self, *args, **kwds):
        DockArea.addDock(self, *args, **kwds)
        self.linkPlots()

    def floatDock(self, *args, **kwds):
        ret = DockArea.floatDock(self, *args, **kwds)
        self.linkPlots()

    def addTempArea(self):
        # Hook addDock to make it all work
        ret = DockArea.addTempArea(self)

        def addDockShim(*args, **kwds):
            DockArea.addDock(ret, *args, **kwds)

            self.linkPlots()

        ret.addDock = addDockShim

        return ret

    ## PySide bug: We need to explicitly redefine these methods
    ## or else drag/drop events will not be delivered.
    def dragEnterEvent(self, *args):
        DockArea.dragEnterEvent(self, *args)

    def dragMoveEvent(self, *args):
        DockArea.dragMoveEvent(self, *args)

    def dragLeaveEvent(self, *args):
        DockArea.dragLeaveEvent(self, *args)

    def dropEvent(self, *args):
        DockArea.dropEvent(self, *args) 

    def linkPlotsRecurse(self, c, x_link=None):
        for i in range(c.count()):
            dock = c.widget(i)

            if isinstance(dock, TContainer):
                x_link = self.linkPlotsRecurse(dock, x_link)

            if not isinstance(dock, Dock):
                continue

            if not dock.name().startswith('TimeSeries'):
                continue

            for w in dock.widgets:
                if isinstance(w, pg.widgets.PlotWidget.PlotWidget):
                    w.setXLink(x_link)

                    x_link = w

        return x_link

    def linkPlots(self):
        # Walk the containers for all vertical containers, find graphs, link
        # ones that are on top of each other in a single container.

        containers, docks = self.findAll()

        for c in containers:
            if isinstance(c, VContainer):
                self.linkPlotsRecurse(c)
            



"""
Implementation of UAV Object Types

Copyright (C) 2014-2015 Tau Labs, http://taulabs.org
Copyright (C) 2015 dRonin, http://dronin.org

Licensed under the GNU LGPL version 2.1 or any later version (see COPYING.LESSER)
"""

import re
from collections import namedtuple, OrderedDict
import xml.etree.ElementTree as etree
from xml.etree.ElementTree import Element, SubElement
import time

try:
    from struct import Struct, calcsize
except:
    from .structshim import Struct, calcsize

RE_SPECIAL_CHARS = re.compile('[\\.\\-\\s\\+/\\(\\)]')

def flatten(lst):
    result = []
    for element in lst:
        if hasattr(element, '__iter__'):
            result.extend(flatten(element))
        else:
            result.append(element)
    return result


class UAVTupleClass():
    """ This is the prototype for a class that contains a uav object. """

    @classmethod
    def _make_to_send(cls, *args, **kwargs):
        """ Accepts all the uavo fields and creates an object.

        The object is dated as of the current time, and has the name and id
        fields set properly.
        """

        return cls(cls._name, round(time.time() * 1000), cls._id, *args, **kwargs)

    def to_bytes(self):
        """ Serializes this object into a byte stream. """
        return self._packstruct.pack(*flatten(self[3:]))

    @classmethod
    def get_size_of_data(cls):
        return cls._packstruct.size

    @classmethod
    def from_bytes(cls, data, timestamp, instance_id, offset=0):
        """ Deserializes and creates an instance of this object.

         - data: the data to deserialize
         - timestamp: the timestamp to put on the object instance
         - offset: an optional index into data where to begin deserialization
        """
        unpack_field_values = cls._packstruct.unpack_from(data, offset)

        field_values = []
        field_values.append(cls._name)

        if timestamp is not None:
            field_values.append(timestamp / 1000.0)

        field_values.append(cls._id)

        if instance_id is not None:
            field_values.append(instance_id)

        # add the remaining fields.  If the thing should be nested, construct
        # an appropriate tuple.
        if not cls._flat:
            pos = 0

            for n in cls._num_subelems:
                if n == 1:
                    field_values.append(unpack_field_values[pos])
                else:
                    field_values.append(tuple(unpack_field_values[pos:pos+n]))
                pos += n

            field_values = tuple(field_values)
        else:
            # Short cut; nothing is nested
            field_values = tuple(field_values) + tuple(unpack_field_values)

        return cls._make(field_values)

    def __repr__(self):
        """ String representation of the contents """
        rep = self.__class__.__name__ + '('
        for field in self._fields:
            field_value = getattr(self, field)

            # Show ENUMs in a fancy way (string and value)
            if hasattr(self, 'ENUMR_' + field):
                if not isinstance(field_value, tuple):
                    field_value = (field_value,)
                value_str = ''
                this_enum = getattr(self, 'ENUMR_' + field)
                for ii, v in enumerate(field_value):
                    value_str += '%s(%d)' % (this_enum.get(v, "UNKNOWN"), v)
                    if ii < len(field_value) - 1:
                        value_str += ', '
                if len(field_value) > 1:
                    value_str = '(%s)' % value_str
            elif field == 'uavo_id':
                value_str = '0x%X' % field_value
            else:
                value_str = str(field_value)
            rep += '%s=%s, ' % (field, value_str)
        rep = rep[:-2] + ')'
        return rep

    def elem_to_string(self, field_name, value):
        mapping = getattr(self, 'ENUMR_' + field_name, None)

        if mapping is None:
            return str(value)

        val = mapping.get(value, None)

        if val is not None:
            return mapping[value]
        else:
            return 'Unknown'

    def to_xml_elem(self):
        munged_name = self._name[5:]
        munged_id = '0x%X' % self._id
        xmlobj = Element('object', { 'name' : munged_name,
            'id' : munged_id })

        raw_dict = self._asdict()

        for field_name in sorted(raw_dict.keys()):
            if field_name == 'name': continue
            if field_name == 'time': continue
            if field_name == 'uavo_id': continue

            field_value = raw_dict[field_name]

            if isinstance(field_value, tuple) or isinstance(field_value, list):
                text_value = ','.join( [ self.elem_to_string(field_name, v) for v in field_value ] )
            else:
                text_value = self.elem_to_string(field_name, field_value)

            xmlfield = SubElement(xmlobj, 'field', { 'name' : field_name, 'values' : text_value })

        return xmlobj


type_enum_map = {
    'int8'    : 0,
    'int16'   : 1,
    'int32'   : 2,
    'uint8'   : 3,
    'uint16'  : 4,
    'uint32'  : 5,
    'float'   : 6,
    'enum'    : 7,
    }

type_numpy_map = {
    'int8'    : 'int8',
    'int16'   : 'int16',
    'int32'   : 'int32',
    'uint8'   : 'uint8',
    'uint16'  : 'uint16',
    'uint32'  : 'uint32',
    'float'   : 'float',
    'enum'    : 'uint8',
    }

struct_element_map = {
    'int8'    : 'b',
    'int16'   : 'h',
    'int32'   : 'i',
    'uint8'   : 'B',
    'uint16'  : 'H',
    'uint32'  : 'I',
    'float'   : 'f',
    'enum'    : 'B',
    }

# This is a very long, scary method.  It parses an XML file describing
# a UAVO and builds an implementation class.
def make_class(collection, xml_file, update_globals=True):
    fields = []

    ##### PARSE THE XML FILE INTO INTERNAL REPRESENTATIONS #####

    tree = etree.fromstring(xml_file)

    subs = {
        'tree'            : tree,
        'object'          : tree.find('object'),
        'fields'          : tree.findall('object/field'),
        'description'     : tree.find('object/description'),
        'access'          : tree.find('object/access'),
        'logging'         : tree.find('object/logging'),
        'telemetrygcs'    : tree.find('object/telemetrygcs'),
        'telemetryflight' : tree.find('object/telemetryflight'),
        }

    name = subs['object'].get('name')
    is_single_inst = int((subs['object'].get('singleinstance') == 'true'))
    is_settings = int((subs['object'].get('settings') == 'true'))

    description = subs['description'].text

    ##### CONSTRUCT PROPER INTERNAL REPRESENTATION OF FIELD DATA #####
    for field in subs['fields']:
        info = {}
        # process typical attributes
        for attr in ['name', 'units', 'type', 'elements', 'elementnames',
                     'parent', 'defaultvalue']:
            info[attr] = field.get(attr)
        if field.get('cloneof'):
            # this is a clone of another field, find its data
            cloneof_name = field.get('cloneof')
            for i, field in enumerate(fields):
                if field['name'] == cloneof_name:
                    clone_info = field.copy()
                    break

            # replace it with the new name
            clone_info['name'] = info['name']
            # use the expanded/substituted info instead of the stub
            info = clone_info
        else:
            if info['elements'] is not None:
                # we've got an inline "elements" attribute
                info['elementnames'] = []
                info['elements'] = int(field.get('elements'))
            elif info['elementnames'] is not None:
                # we've got an inline "elementnames" attribute
                info['elementnames'] = []
                info['elements'] = 0
                for elementname in field.get('elementnames').split(','):
                    info['elementnames'].append(RE_SPECIAL_CHARS.sub('', elementname))
                    info['elements'] += 1
            else:
                # we must have one or more elementnames/elementname elements in this sub-tree
                info['elementnames'] = []
                info['elements'] = 0
                for elementname_text in [elementname.text for elementname in field.findall('elementnames/elementname')]:
                    info['elementnames'].append(elementname_text)
                    info['elements'] += 1

            if info['type'] == 'enum':
                info['options'] = OrderedDict()
                if field.get('options'):
                    # we've got an inline "options" attribute
                    for ii, option_text in enumerate(field.get('options').split(',')):
                        info['options'][option_text.strip()] = ii
                else:
                    # we must have some 'option' elements in this sub-tree
                    for ii, option_text in enumerate([option.text
                            for option in field.findall('options/option')]):
                        info['options'][option_text.strip()] = ii

            # convert type string to an int
            info['type_val'] = type_enum_map[info['type']]

            # Get parent
            if info['parent'] is not None:
                parent_name, field_name = info['parent'].split('.')
                parent_class = collection.find_by_name(parent_name)
                if len(info['options']) == 0:
                    parent_options = getattr(parent_class, 'ENUMR_' + field_name)
                    for k, v in sorted(iter(parent_options.items()), key=lambda x: x[0]):
                        info['options'][v] = k
                else:
                    parent_options = getattr(parent_class, 'ENUM_' + field_name)
                    for k in info['options'].keys():
                        info['options'][k] = parent_options[k]

            # Add default values
            if info['defaultvalue'] is not None:
                if info['type'] == 'enum':
                    try:
                        values = tuple(info['options'][v.strip()]
                                       for v in info['defaultvalue'].split(','))
                    except KeyError:
                        print('Invalid default value: %s.%s has no option %s' % (name, info['name'], info['defaultvalue']))
                        values = (0,)
                else:  # float or int
                    values = tuple(float(v) for v in info['defaultvalue'].split(','))
                    if info['type'] != float:
                        values = tuple(int(v) for v in values)

                if len(values) == 1:
                    values = values[0]
                info['defaultvalue'] = values
            else:
                # Use 0 as the default
                info['defaultvalue'] = 0

            if info['elements'] > 1 and not isinstance(info['defaultvalue'], tuple):
                info['defaultvalue'] = (info['defaultvalue'],) * info['elements']

            if field.find('description') != None:
                info['description'] = field.find('description').text

        fields.append(info)

    # Sort fields by size (bigger to smaller) to ensure alignment when packed
    fields.sort(key=lambda x: calcsize(struct_element_map[x['type']]), reverse = True)

    ##### CALCULATE THE APPROPRIATE UAVO ID #####
    hash_calc = UAVOHash()

    hash_calc.update_hash_string(name)
    hash_calc.update_hash_byte(is_settings)
    hash_calc.update_hash_byte(is_single_inst)

    for field in fields:
        hash_calc.update_hash_string(field['name'])
        hash_calc.update_hash_byte(field['elements'])
        hash_calc.update_hash_byte(field['type_val'])
        if field['type'] == 'enum':
            next_idx = 0
            for option, idx in field['options'].items():
                if idx != next_idx:
                    hash_calc.update_hash_byte(idx)
                hash_calc.update_hash_string(option)
                next_idx = idx + 1

    uavo_id = hash_calc.get_hash()

    ##### FORM A STRUCT TO PACK/UNPACK THIS UAVO'S CONTENT #####
    formats = []
    num_subelems = []

    is_flat = True

    # add formats for each field
    for f in fields:
        if f['elements'] != 1:
            is_flat = False

        num_subelems.append(f['elements'])

        formats.append('' + f['elements'].__str__() + struct_element_map[f['type']])

    fmt = Struct('<' + ''.join(formats))

    ##### CALCULATE THE NUMPY TYPE ASSOCIATED WITH THIS CLASS #####
    dtype  = [('name', 'S20'), ('time', 'double'), ('uavo_id', 'uint')]

    if not is_single_inst:
        dtype += ('inst_id', 'uint'),

    for f in fields:
        if f['elements'] != 1:
            dtype += [(f['name'], '(' + repr(f['elements']) + ",)" + type_numpy_map[f['type']])]
        else:
            dtype += [(f['name'], type_numpy_map[f['type']])]

    ##### DYNAMICALLY CREATE A CLASS TO CONTAIN THIS OBJECT #####
    tuple_fields = ['name', 'time', 'uavo_id']
    if not is_single_inst:
        tuple_fields.append("inst_id")

    tuple_fields.extend([f['name'] for f in fields])

    name = 'UAVO_' + name

    class tmpClass(UAVTupleClass, namedtuple(name, tuple_fields)):
        _packstruct = fmt
        _flat = is_flat
        _name = name
        _id = uavo_id
        _single = is_single_inst
        _num_subelems = num_subelems
        _dtype = dtype
        _is_settings = is_settings
        _units = {f['name'] : f['units'] for f in fields}

    # This is magic for two reasons.  First, we create the class to have
    # the proper dynamic name.  Second, we override __slots__, so that
    # child classes don't get a dict / keep all their namedtuple goodness
    tuple_class = type(name, (tmpClass,), { "__slots__" : () })

    # Set the default values for the constructor
    defaults = [name, 0, uavo_id]
    defaults.extend(f['defaultvalue'] for f in fields)
    tuple_class.__new__.__defaults__ = tuple(defaults)

    # Add enums
    for field in fields:
        if field['type'] == 'enum':
            enum = field['options']
            mapping = dict((key, value) for key, value in enum.items())
            reverse_mapping = dict((value, key) for key, value in enum.items())
            setattr(tuple_class, 'ENUM_' + field['name'], mapping)
            setattr(tuple_class, 'ENUMR_' + field['name'], reverse_mapping)

    if update_globals:
        globals()[tuple_class.__name__] = tuple_class


    return tuple_class


class UAVOHash():
    def __init__(self):
        self.hval = 0

    def update_hash_byte(self, value):
        self.hval = (self.hval ^ ((self.hval << 5) + (self.hval >> 2) + value)) & 0x0FFFFFFFF

    def update_hash_string(self, string):
        for c in string:
            self.update_hash_byte(ord(c))

    def get_hash(self):
        return self.hval & 0x0FFFFFFFE



"""
UAVO collection interface.

Copyright (C) 2014-2015 Tau Labs, http://taulabs.org
Copyright (C) 2015 dRonin, http://dronin.org

Licensed under the GNU LGPL version 2.1 or any later version (see COPYING.LESSER)
"""

from . import uavo

import operator
import os.path as op

GITHASH_OF_LAST_RESORT = 'Release-20160120.3'

class UAVOCollection(dict):
    def __init__(self):
        self.clear()

    def find_by_name(self, uavo_name):
        if uavo_name[0:5]!='UAVO_':
            uavo_name = 'UAVO_' + uavo_name

        for u in self.values():
            if u._name == uavo_name:
                return u

        return None

    def get_settings_objects(self):
        objs = [ u for u in self.values() if u._is_settings ]
        objs.sort(key=operator.attrgetter('_name'))

        return objs

    def from_file_contents(self, content_list):
        some_processed = True

        # There are dependencies here
        while some_processed:
            some_processed = False

            unprocessed = []

            # Build up the UAV objects from the xml definitions
            for contents in content_list:
                try:
                    u = uavo.make_class(self, contents)

                    # add this uavo definition to our dictionary
                    self.update([('{0:08x}'.format(u._id), u)])

                    some_processed = True
                except Exception:
                    unprocessed.append(contents)

            content_list = unprocessed

        if len(content_list):
            raise Exception("Unable to parse some uavo files")

    def from_tar_file(self, t):
        # Get the file members
        content_list = []

        for f_info in t.getmembers():
            if f_info.name.count("oplinksettings") > 0:
                continue

            if not f_info.name.endswith('.xml'):
                continue

            f = t.extractfile(f_info)
            content_list.append(f.read())
            f.close()

        self.from_file_contents(content_list)

    def from_tar_bytes(self, contents):
        from io import BytesIO
        import tarfile

        # coerce the tar file data into a file object so that tarfile likes it
        fobj = BytesIO(contents)

        # feed the tar file data to a tarfile object
        with tarfile.open(fileobj=fobj) as t:
            self.from_tar_file(t)

    def from_git_hash(self, githashes):
        if not isinstance(githashes, list):
            githashes = [ githashes ]

        for h in githashes:
            try:
                #
                # Grab the exact uavo definition files from the git repo using 
                # the provided hash
                #
                import subprocess

                # Get the directory where the code is located
                src_dir = op.join(op.dirname(__file__), "..", "..")

                p = subprocess.Popen(['git', 'archive', h, '--', 'shared/uavobjectdefinition/'],
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                     cwd=src_dir)
                # grab the tar file data
                git_archive_data, git_archive_errors = p.communicate()

                if p.returncode == 0:
                    self.from_tar_bytes(git_archive_data)
                    return
            except Exception:
                # Popen isn't available on GAE, so all of the above fails.
                pass

            try:
                from six.moves.urllib.request import urlopen

                # TODO -- this can be bundled into a single request and unrolled.
                web_data = urlopen("http://dronin-autotown.appspot.com/uavos/%s" % (h))

                if web_data.getcode() != 200:
                    continue

                self.from_tar_bytes(web_data.read())
                return
            except Exception:
                pass

        raise Exception("Couldn't get any UAVO definitions by githash")

    def from_uavo_xml_path(self, path):
        import os
        import glob

        content_list = []

        for file_name in glob.glob(os.path.join(path, '*.xml')):
            with open(file_name, 'rU') as f:
                content_list.append(f.read())

        self.from_file_contents(content_list)



"""
Crude implementation of struct.Struct, intended for pyjs use

Copyright (C) 2016 dRonin, http://dronin.org

Licensed under the GNU LGPL version 2.1 or any later version (see COPYING.LESSER)
"""

import struct

def calcsize(fmt_string):
    return struct.calcsize(fmt_string)

class Struct:
    def __init__(self, fmt_string):
        self.fmt_string = fmt_string
        self.size = calcsize(fmt_string)

    def pack(self, *args):
        return struct.pack(self.fmt_string, *args)

    def unpack(self, buff):
        return struct.unpack(self.fmt_string, buff)

    def unpack_from(self, buff, offset):
        return struct.unpack_from(self.fmt_string, buff, offset)



"""
Interface to telemetry streams -- log, network, or serial.

Copyright (C) 2014-2015 Tau Labs, http://taulabs.org
Copyright (C) 2015-2016 dRonin, http://dronin.org
Licensed under the GNU LGPL version 2.1 or any later version (see COPYING.LESSER)
"""

import socket
import time
import errno

from . import uavtalk, uavo_collection, uavo

import os

from abc import ABCMeta, abstractmethod

from six import with_metaclass

class TelemetryBase(with_metaclass(ABCMeta)):
    """
    Basic (abstract) implementation of telemetry used by all stream types.
    """

    def __init__(self, githash=None, service_in_iter=True,
            iter_blocks=True, use_walltime=True, do_handshaking=False,
            gcs_timestamps=False, name=None, progress_callback=None):

        """Instantiates a telemetry instance.  Called only by derived classes.
         - githash: revision control id of the UAVO's used to communicate.
             if unspecified, we use the version in this source tree.
         - service_in_iter: whether the actual iterator should service the
             connection/read the file.  If you don't specify this, you probably
             want to call start_thread()
         - iter_blocks: whether the iterator should block for new data.  Only
             really meaningful for network/serial
         - use_walltime: if true, automatically place current time into packets
             received.
         - do_handshaking: if true, indicates this is an interactive stream
             where we should speak the UAVO_GCSTelemetryStats connection status
             protocol.
         - gcs_timestamps: if true, this means we are reading from a file with
             the GCS timestamp protocol.  if None, request autodetection
         - name: a filename to store into .filename for legacy purposes
         - progress_callback: a function to call periodically with progress
             information
        """

        uavo_defs = uavo_collection.UAVOCollection()

        if githash:
            uavo_defs.from_git_hash(githash)
        else:
            xml_path = os.path.join(os.path.dirname(__file__), "..", "..",
                                    "shared", "uavobjectdefinition")
            uavo_defs.from_uavo_xml_path(xml_path)

        self.githash = githash

        self.uavo_defs = uavo_defs
        self.uavtalk_generator = uavtalk.process_stream(uavo_defs,
            use_walltime=use_walltime, gcs_timestamps=gcs_timestamps,
            progress_callback=progress_callback)

        self.uavtalk_generator.send(None)

        self.uavo_list = []

        self.last_values = {}

        from threading import Condition
        self.cond = Condition()

        self.service_in_iter = service_in_iter
        self.iter_blocks = iter_blocks

        self.do_handshaking = do_handshaking
        self.filename = name

        self.eof = False

        self.first_handshake_needed = self.do_handshaking

        if do_handshaking:
            self.GCSTelemetryStats = uavo_defs.find_by_name('UAVO_GCSTelemetryStats')
            self.FlightTelemetryStats = uavo_defs.find_by_name('UAVO_FlightTelemetryStats')

    def as_numpy_array(self, match_class, filter_cond=None):
        """ Transforms all received instances of a given object to a numpy array.

        match_class: the UAVO_* class you'd like to match.
        """

        import numpy as np

        # Find the subset of this list that is of the requested class
        filtered_list = [x for x in self if isinstance(x, match_class)]

        # Perform any additional requested filtering
        if filter_cond is not None:
            filtered_list = list(filter(filter_cond, filtered_list))

        # Check for an empty list
        if filtered_list == []:
            return np.array([])

        return np.array(filtered_list, dtype=match_class._dtype)

    def __iter__(self):
        """ Iterator service routine. """
        iterIdx = 0

        self.cond.acquire()

        while True:
            if iterIdx < len(self.uavo_list):
                obj = self.uavo_list[iterIdx]

                iterIdx += 1

                self.cond.release()

                yield obj

                self.cond.acquire()
            elif self.iter_blocks and not self._done():
                if self.service_in_iter:
                    self.cond.release()

                    self.service_connection()

                    self.cond.acquire()
                else:
                    # wait for another thread to fill it in
                    self.cond.wait()
            else:
                # Don't really recommend this mode anymore/maybe remove
                if self.service_in_iter and not self._done():
                    # Do at least one non-blocking attempt
                    self.cond.release()

                    self.service_connection(0)

                    self.cond.acquire()

                # I think this should probably keep the index so that
                # new iterations pick up where we were.. XXX TODO
                # takes some thought as to what is "right"
                if iterIdx >= len(self.uavo_list):
                    break

    def __make_handshake(self, handshake):
        return self.GCSTelemetryStats._make_to_send(
                Status=self.GCSTelemetryStats.ENUM_Status[handshake])

    def send_object(self, obj):
        self._send(uavtalk.send_object(obj))

    def __handle_handshake(self, obj):
        if obj.name == "UAVO_FlightTelemetryStats":
            # Handle the telemetry handshaking

            if obj.Status == obj.ENUM_Status['Disconnected']:
                # Request handshake
                print("Disconnected")
                send_obj = self.__make_handshake('HandshakeReq')
            elif obj.Status == obj.ENUM_Status['HandshakeAck']:
                # Say connected
                print("Handshake ackd")
                send_obj = self.__make_handshake('Connected')
            elif obj.Status == obj.ENUM_Status['Connected']:
                print("Connected")
                send_obj = self.__make_handshake('Connected')

            self.send_object(send_obj)

    def request_object(self, obj):
        if not self.do_handshaking:
            raise ValueError("Can only request on handshaking/bidir sessions")

        self._send(uavtalk.request_object(obj))

    def __handle_frames(self, frames):
        objs = []

        if frames == b'':
            self.eof=True
        else:
            obj = self.uavtalk_generator.send(frames)

            while obj:
                if self.do_handshaking:
                    self.__handle_handshake(obj)

                objs.append(obj)

                obj = self.uavtalk_generator.send(b'')

        # Only traverse the lock when we've processed everything in this
        # batch.
        with self.cond:
            # keep everything in ram forever
            # for now-- in case we wanna see
            self.uavo_list.extend(objs)

            for obj in objs:
                self.last_values[obj.__class__]=obj

            self.cond.notifyAll()

    def get_last_values(self):
        """ Returns the last instance of each kind of object received. """
        with self.cond:
            return self.last_values.copy()

    def wait_connection(self):
        """ Waits for the connection handshaking to complete. """

        with self.cond:
            while True:
                fts = self.last_values.get(self.FlightTelemetryStats)

                if fts is not None:
                    if fts.Status == fts.ENUM_Status['Connected']:
                        return

                self.cond.wait()

    def start_thread(self):
        """ Starts a separate thread to service this telemetry connection. """
        if self.service_in_iter:
            # TODO sane exceptions here.
            raise

        if self._done():
            raise

        from threading import Thread

        def run():
            while not self._done():
                self.service_connection()

        t = Thread(target=run, name="telemetry svc thread")

        t.daemon=True

        t.start()

    def service_connection(self, timeout=None):
        """
        Receive and parse data from a connection and handle the basic
        handshaking with the flight controller
        """

        if timeout is not None:
            finish_time = time.time() + timeout
        else:
            finish_time = None

        if self.first_handshake_needed:
            send_obj = self.__make_handshake('Disconnected')
            self.send_object(send_obj)

            self.first_handshake_needed = False
        data = self._receive(finish_time)
        self.__handle_frames(data)

    @abstractmethod
    def _receive(self, finish_time):
        return

    # No implementation required, so not abstract
    def _send(self, msg):
        return

    def _done(self):
        with self.cond:
            return self.eof

class BidirTelemetry(TelemetryBase):
    """
    Stuff involved in buffering for bidirection telemetry.
    """
    def __init__(self, *args, **kwargs):
        """ Abstract instantiation of bidir telemetry

        Should only be called by derived classes.

        Meaningful parameters passed up to TelemetryBase include: githash,
        service_in_iter, iter_blocks, use_walltime
        """
        TelemetryBase.__init__(self, do_handshaking=True,
                gcs_timestamps=False,  *args, **kwargs)

        self.recv_buf = b''
        self.send_buf = b''

    def _receive(self, finish_time):
        """ Fetch available data from file descriptor. """

        # Always do some minimal IO if possible
        self._do_io(0)

        while (len(self.recv_buf) < 1) and self._do_io(finish_time):
            pass

        if len(self.recv_buf) < 1:
            return None

        ret = self.recv_buf
        self.recv_buf = b''

        return ret

    def _send(self, msg):
        """ Send a string to the controller """

        self.send_buf += msg

        self._do_io(0)

    @abstractmethod
    def _do_io(self, finish_time):
        return


class FDTelemetry(BidirTelemetry):
    """
    Implementation of bidirectional telemetry from a file descriptor.

    Intended for network streams.
    """

    def __init__(self, fd, *args, **kwargs):
        """ Instantiates a telemetry instance on a given fd.

        Probably should only be called by derived classes.

         - fd: the file descriptor to perform telemetry operations upon

        Meaningful parameters passed up to TelemetryBase include: githash,
        service_in_iter, iter_blocks, use_walltime
        """

        BidirTelemetry.__init__(self, *args, **kwargs)

        self.fd = fd

    # Call select and do one set of IO operations.
    def _do_io(self, finish_time):
        import select

        rdSet = []
        wrSet = []

        did_stuff = False

        if len(self.recv_buf) < 1024:
            rdSet.append(self.fd)

        if len(self.send_buf) > 0:
            wrSet.append(self.fd)

        now = time.time()
        if finish_time is None:
            r,w,e = select.select(rdSet, wrSet, [])
        else:
            tm = finish_time-now
            if tm < 0: tm=0

            r,w,e = select.select(rdSet, wrSet, [], tm)

        if r:
            # Shouldn't throw an exception-- they just told us
            # it was ready for read.
            # TODO: Figure out why read sometimes fails when using sockets
            try:
                chunk = os.read(self.fd, 1024)
                if chunk == '':
                    raise RuntimeError("stream closed")

                self.recv_buf = self.recv_buf + chunk

                did_stuff = True
            except OSError as err:
                # For some reason, we sometimes get a
                # "Resource temporarily unavailable" error
                if err.errno != errno.EAGAIN:
                    raise

        if w:
            written = os.write(self.fd, self.send_buf)

            if written > 0:
                self.send_buf = self.send_buf[written:]

            did_stuff = True

        return did_stuff

class NetworkTelemetry(FDTelemetry):
    """ TCP telemetry interface. """
    def __init__(self, host="127.0.0.1", port=9000, *args, **kwargs):
        """ Creates a telemetry instance talking over TCP.

         - host: hostname to connect to (default localhost)
         - port: port number to communicate on (default 9000)

        Meaningful parameters passed up to TelemetryBase include: githash,
        service_in_iter, iter_blocks, use_walltime
        """

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))

        s.setblocking(0)

        self.sock = s

        FDTelemetry.__init__(self, fd=s.fileno(), *args, **kwargs)

# TODO XXX : Plumb appropriate cleanup / file close for these classes

class SerialTelemetry(BidirTelemetry):
    """ Serial telemetry interface """
    def __init__(self, port, speed=115200, *args, **kwargs):
        """ Creates telemetry instance talking over (real or virtual) serial port.

         - port: Serial port path
         - speed: Baud rate (doesn't really matter for VCP, defaults 115200)

        Meaningful parameters passed up to TelemetryBase include: githash,
        service_in_iter, iter_blocks, use_walltime

        Requires the module pyserial to provide OS independance.
        """

        import serial

        self.ser = serial.Serial(port, speed, timeout=0, writeTimeout=0.001)

        BidirTelemetry.__init__(self, *args, **kwargs)

    # Call select and do one set of IO operations.
    def _do_io(self, finish_time):
        import serial

        did_stuff = False

        while not did_stuff:
            try:
                chunk = self.ser.read(1024)

                if chunk != '':
                    did_stuff = True
                    self.recv_buf = self.recv_buf + chunk
            except serial.serialutil.SerialException:
                # Ignore this; looks like a pyserial bug
                pass

            if self.send_buf != '':
                try:
                    written = self.ser.write(self.send_buf)

                    if written > 0:
                        self.send_buf = self.send_buf[written:]
                        did_stuff = True
                except serial.SerialTimeoutException:
                    pass

            now = time.time()
            if finish_time is not None:
                if now > finish_time:
                    break

            time.sleep(0.0025)    # 2.5ms, ~28 bytes of time at 115200

        return did_stuff

class FileTelemetry(TelemetryBase):
    """ Telemetry interface to data in a file """

    def __init__(self, file_obj, parse_header=False,
             *args, **kwargs):
        """ Instantiates a telemetry instance reading from a file.

         - file_obj: the file object to read from
         - parse_header: whether to read a header like the GCS writes from the
           file.

        Meaningful parameters passed up to TelemetryBase include: githash,
        service_in_iter, iter_blocks, gcs_timestamps
        """

        self.f = file_obj

        if parse_header:
            # Check the header signature
            #    First line is "dRonin git hash:" or "Tau Labs git hash:"
            #    Second line is the actual git hash
            #    Third line is the UAVO hash
            #    Fourth line is "##" (only from GCS)

            # Scan up to 100 "lines" looking for the signature, in case
            # there's garbage at the beginning of the log
            found = False

            for i in range(100):
                sig = self.f.readline()
                if sig.endswith(b'dRonin git hash:\n') or sig.endswith(b'Tau Labs git hash:\n'):
                    found = True
                    break;

            if not found:
                print("Source file does not have a recognized header signature")
                raise IOError("no header signature")

            # Determine the git hash that this log file is based on
            githash = self.f.readline()[:-1]
            if githash.find(b':') != -1:
                import re
                githash = re.search(b':(\w*)\W', githash).group(1)

            # For python3, convert from byte string.
            githash = githash.decode('latin-1')

            print("Log file is based on git hash: %s" % githash)

            uavohash = self.f.readline()
            # divider only occurs on GCS-type streams.  This causes us to
            # miss first objects in telemetry-type streams
            # divider = self.f.readline()

            TelemetryBase.__init__(self, iter_blocks=True,
                do_handshaking=False, githash=githash, use_walltime=False,
                *args, **kwargs)
        else:
            TelemetryBase.__init__(self, iter_blocks=True,
                do_handshaking=False, use_walltime=False, *args, **kwargs)

        self.done=False

    def _receive(self, finish_time):
        """ Fetch available data from file """

        buf = self.f.read(524288)   # 512k

        return buf

def get_telemetry_by_args(desc="Process telemetry", service_in_iter=True,
        iter_blocks=True):
    """ Parses command line to decide how to get a telemetry object. """
    # Setup the command line arguments.
    import argparse
    parser = argparse.ArgumentParser(description=desc)

    # Log format indicates this log is using the old file format which
    # embeds the timestamping information between the UAVTalk packet
    # instead of as part of the packet
    parser.add_argument("-t", "--timestamped",
                        action  = 'store_false',
                        default = None,
                        help    = "indicate that this is not timestamped in GCS format")

    parser.add_argument("-g", "--githash",
                        action  = "store",
                        dest    = "githash",
                        help    = "override githash for UAVO XML definitions")

    parser.add_argument("-s", "--serial",
                        action  = "store_true",
                        default = False,
                        dest    = "serial",
                        help    = "indicates that source is a serial port")

    parser.add_argument("-b", "--baudrate",
                        action  = "store",
                        dest    = "baud",
                        default = "115200",
                        help    = "baud rate for serial communications")

    parser.add_argument("source",
                        help  = "file, host:port, or serial port to get telemetry from")

    # Parse the command-line.
    args = parser.parse_args()

    parse_header = False
    githash = None

    if args.githash is not None:
        # If we specify the log header no need to attempt to parse it
        githash = args.githash
    else:
        parse_header = True # only for files

    from dronin import telemetry

    if args.serial:
        return telemetry.SerialTelemetry(args.source, speed=args.baud,
                service_in_iter=service_in_iter, iter_blocks=iter_blocks,
                githash=githash)

    if args.baud != "115200":
        parser.print_help()
        raise ValueError("Baud rates only apply to serial ports")

    import os.path

    if os.path.isfile(args.source):
        file_obj = open(args.source, 'rb')

        if parse_header:
            t = telemetry.FileTelemetry(file_obj, parse_header=True,
                gcs_timestamps=args.timestamped, name=args.source)
        else:
            t = telemetry.FileTelemetry(file_obj, parse_header=False,
                gcs_timestamps=args.timestamped, name=args.source,
                githash=githash)

        return t

    # OK, running out of options, time to try the network!
    host,sep,port = args.source.partition(':')

    if sep != ':':
        parser.print_help()
        raise ValueError("Target doesn't exist and isn't a network address")

    return telemetry.NetworkTelemetry(host=host, port=int(port), name=args.source,
            service_in_iter=service_in_iter, iter_blocks=iter_blocks,
            githash=githash)



 # Use 3.x import model.

#-------------------------------------------------------------------------------
__all__ = ()

from . import logfs
from . import telemetry
from . import uavo
from . import uavo_collection
from . import uavtalk



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

long_description = """
This is the dRonin UAVTalk API.  With these modules, it is possible to
communicate with dRonin flight controllers and interpret log files and
settings partitions.
"""

setup(
    name='dronin',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='20160825',

    description='dRonin Python API',
    long_description=long_description,

    # The project's main homepage.
    url='http://dronin.org/',

    # Author details
    author='dronin',
    author_email='info@dronin.org',

    # Choose your license
    license='LGPLv2+',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Embedded Systems',
        "Topic :: Software Development :: Object Brokering",


        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],

    # What does your project relate to?
    keywords='dronin flight controller quadcopter',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages = ['dronin', 'dronin.logviewer'],

    # Just requires the base python system to run
    install_requires=['six'],

    # 'all' target requires python3 because there are only PyQt5 wheels for
    # python3.
    extras_require={
        'all': ['pyserial', 'numpy', 'matplotlib', 'dronin-pyqtgraph', 'PyQt5'],
    },

    scripts = [ 'dronin-dumplog', 'dronin-halt',
        'dronin-getconfig', 'dronin-logfsimport',
        'dronin-shell' ],
#    package_data={
#        'sample': ['package_data.dat'],
#    },

#    data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'gui_scripts': [
            'dronin-logview=dronin.logviewer.logviewer:main',
        ],
    },
)



"""
Implements the uavtalk protocol.

Copyright (C) 2014-2015 Tau Labs, http://taulabs.org
Copyright (C) 2015 dRonin, http://dronin.org

Licensed under the GNU LGPL version 2.1 or any later version (see COPYING.LESSER)

Ordinarily one would use the methods exposed by the telemetry module instead of
this interface.
"""

try:
    from struct import Struct
except:
    from .structshim import Struct

import time

__all__ = [ "send_object", "process_stream" ]

from six import int2byte, indexbytes, byte2int, iterbytes

# Constants used for UAVTalk parsing
(MIN_HEADER_LENGTH, MAX_HEADER_LENGTH, MAX_PAYLOAD_LENGTH) = (8, 12, (256-12))
(SYNC_VAL) = (0x3C)
(TYPE_MASK, TYPE_VER) = (0x78, 0x20)
(TIMESTAMPED) = (0x80)
(TYPE_OBJ, TYPE_OBJ_REQ, TYPE_OBJ_ACK, TYPE_ACK, TYPE_NACK, TYPE_OBJ_TS, TYPE_OBJ_ACK_TS) = (0x00, 0x01, 0x02, 0x03, 0x04, 0x80, 0x82)

# Serialization of header elements

# sync(1) + type(1) + len(2) + objid(4)
header_fmt = Struct("<BBHL")
logheader_fmt = Struct("<IQ")
timestamp_fmt = Struct("<H")
instance_fmt = Struct("<H")

# CRC lookup table
crc_table = [
    0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15, 0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
    0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65, 0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
    0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5, 0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
    0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85, 0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
    0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2, 0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
    0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2, 0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
    0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32, 0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
    0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42, 0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
    0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c, 0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
    0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec, 0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
    0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c, 0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
    0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c, 0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
    0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b, 0x76, 0x71, 0x78, 0x7f, 0x6a, 0x6d, 0x64, 0x63,
    0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b, 0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
    0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb, 0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8d, 0x84, 0x83,
    0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb, 0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3
]

def process_stream(uavo_defs, use_walltime=False, gcs_timestamps=None,
        progress_callback=None):
    """Generator function that parses uavotalk stream.
    
    You are expected to send more bytes, or '' to it, until EOF.  Then send
    None.  After that, you may continue to receive objects back because of
    buffering."""

    # These are used for accounting for timestamp wraparound
    timestamp_base = 0
    last_timestamp = 0

    received = 0

    buf = b''
    buf_offset = 0

    past_bytes = 0

    pending_pieces = []

    while True:
        # If we don't have sufficient data buffered, join up any chunks we've 
        # been given to ensure pending_pieces is empty for the rest of this loop.
        #
        # in other words, don't mix the pending_pieces drain model and the
        # buffer concatenation model within a loop iteration.
        #
        # 10k chosen here to be bigger than any plausible uavo; could calculate
        # this instead from our known uav objects
        if len(buf) - buf_offset < 10240:
            past_bytes += buf_offset

            #print "stitch pp=%d"%(len(pending_pieces))
            pending_pieces.insert(0, buf[buf_offset:])
            buf_offset = 0

            buf = b''.join(pending_pieces)
            pending_pieces = []

        if gcs_timestamps is None or gcs_timestamps == True:
            while len(buf) < (header_fmt.size + logheader_fmt.size + buf_offset):
                rx = yield None

                if rx is None:
                    return

                buf = buf + rx

            overrideTimestamp, logHdrLen = logheader_fmt.unpack_from(buf,buf_offset)

            if gcs_timestamps is None:
                if ((logHdrLen > 1000) or ( overrideTimestamp > 100000000)):
                    if indexbytes(buf, buf_offset) == SYNC_VAL:
                        print("Autodetect: no gcs-type timestamps")
                        gcs_timestamps = False
                    else:
                        print("Autodetect: punting to next cycle")
                        buf_offset += 1
                        continue
                else:
                    if indexbytes(buf, buf_offset + logheader_fmt.size) == SYNC_VAL:
                        print("Autodetect: GCS-type timestamps likely")
                        gcs_timestamps = True
                        buf_offset += logheader_fmt.size
                    else:
                        print("Autodetect: punting to next cycle")
            else:
                buf_offset += logheader_fmt.size

        # Ensure we have enough room for all the
        # plain required fields to avoid duplicating
        # this code lots.
        # sync(1) + type(1) + len(2) + objid(4)

        while (len(buf) < header_fmt.size + buf_offset) or (indexbytes(buf, buf_offset) != SYNC_VAL):
            #print "waitingsync len=%d, offset=%d"%(len(buf), buf_offset)

            if len(buf) < header_fmt.size + 1 + buf_offset:
                rx = yield None

                if rx is None:
                    #end of stream, stopiteration
                    return

                buf = buf + rx

            for i in range(buf_offset, len(buf)):
                if indexbytes(buf, i) == SYNC_VAL:
                    break

            #print "skipping from %d to %d"%(buf_offset, i)

            # Trim off irrelevant stuff, loop and try again
            buf_offset = i

        (sync, pack_type, pack_len, objId) = header_fmt.unpack_from(buf, buf_offset)

        if (pack_type & TYPE_MASK) != TYPE_VER:
            print("badver %x"%(pack_type))
            buf_offset += 1
            continue    # go to top to look for sync
    
        pack_type &= ~ TYPE_MASK
        
        if pack_len < MIN_HEADER_LENGTH or pack_len > MAX_HEADER_LENGTH + MAX_PAYLOAD_LENGTH:
            print("badlen %d"%(pack_len))
            buf_offset += 1
            continue
        
        # Search for object.
        uavo_key = '{0:08x}'.format(objId)
        if not uavo_key in uavo_defs:
            #print "Unknown object 0x" + uavo_key
            obj = None
        else:
            obj = uavo_defs[uavo_key]

        # Determine data length
        if pack_type == TYPE_OBJ_REQ or pack_type == TYPE_ACK or pack_type == TYPE_NACK:
            obj_len = 0
            timestamp_len = 0
            obj = None
        else:
            if obj is not None:
                timestamp_len = timestamp_fmt.size if pack_type == TYPE_OBJ_TS or pack_type == TYPE_OBJ_ACK_TS else 0
                obj_len = obj.get_size_of_data()
            else:
                # we don't know anything, so fudge to keep sync.
                timestamp_len = 0
                obj_len = pack_len - header_fmt.size

        if obj is not None and not obj._single:
            instance_len = 2
        else:
            instance_len = 0

        # Check length and determine next state
        if obj_len >= MAX_PAYLOAD_LENGTH:
            print("bad len-- bad xml?")
            #should never happen; requires invalid uavo xml
            buf_offset += 1
            continue

        # calc_size, AKA timestamp, and obj data
        # as appropriate, plus our current header
        # also equivalent to the offset of the CRC in the packet

        calc_size = header_fmt.size + instance_len + timestamp_len + obj_len

        # Check the lengths match
        if calc_size != pack_len:
            print("mismatched size id=%s %d vs %d, type %d"%(uavo_key,
                calc_size, pack_len, pack_type))

            # packet error - mismatched packet size
            # Consume a byte to try syncing right after where we
            # did...
            buf_offset += 1
            continue

        # OK, at this point we are seriously hoping to receive
        # a packet.  Time for another loop to make sure we have
        # enough data.
        # +1 here is for CRC-8
        while len(buf) < calc_size + 1 + buf_offset:
            rx = yield None

            if rx is None:
                #end of stream, stopiteration
                return

            buf += rx

        # check the CRC byte

        cs = calcCRC(buf[buf_offset:calc_size+buf_offset])
        recv_cs = indexbytes(buf, buf_offset + calc_size)

        if recv_cs != cs:
            print("Bad crc. Got", recv_cs, "but predicted", cs)

            buf_offset += 1

            continue

        if instance_len:
            instance_id = instance_fmt.unpack_from(buf, header_fmt.size + buf_offset)[0]
        else:
            instance_id = None

        if timestamp_len:
            # pull the timestamp from the packet
            timestamp = timestamp_fmt.unpack_from(buf, header_fmt.size + instance_len + buf_offset)[0]

            # handle wraparound
            if timestamp < last_timestamp:
                timestamp_base = timestamp_base + 65536
            last_timestamp = timestamp
            timestamp += timestamp_base
        else:
            timestamp = last_timestamp

        if use_walltime:
            timestamp = int(time.time()*1000.0)

        if gcs_timestamps:
            timestamp = overrideTimestamp

        if obj is not None:
            offset = header_fmt.size + instance_len + timestamp_len + buf_offset
            objInstance = obj.from_bytes(buf, timestamp, instance_id, offset=offset)
            received += 1
            if not (received % 10000):
                if progress_callback is not None:
                    progress_callback(received, past_bytes + buf_offset + calc_size + 1)
                print("received %d objs"%(received))

            next_recv = yield objInstance
        else:
            next_recv = None

        buf_offset += calc_size + 1

        if next_recv is not None and next_recv != '':
            pending_pieces.append(next_recv)

def send_object(obj):
    """Generates a string containing a UAVTalk packet describing this object"""

    hdr = header_fmt.pack(SYNC_VAL, TYPE_OBJ | TYPE_VER,
        header_fmt.size + obj.get_size_of_data(),
        obj._id)

    packet = hdr + obj.to_bytes()

    packet += int2byte(calcCRC(packet))

    return packet

def request_object(obj):
    """Makes a request for this object"""
    packet = header_fmt.pack(SYNC_VAL, TYPE_OBJ_REQ | TYPE_VER,
        header_fmt.size, obj._id)

    packet += int2byte(calcCRC(packet))

    return packet

def calcCRC(s):
    """
    Calculate a CRC consistently with how they are computed on the firmware side
    """

    cs = 0

    for c in iterbytes(s):
        cs = crc_table[cs ^ c]

    return cs
